import ipaddress
import threading
import unittest

from pycrowdsec.client import StreamDecisionClient


class TestStreamDecisionClient(unittest.TestCase):
    def setUp(self):
        self.client = StreamDecisionClient("abcd")

    def test_process_response(self):
        response = {
            "deleted": [
                {
                    "duration": "-40h37m10.022674981s",
                    "id": 1,
                    "origin": "cscli",
                    "scenario": "manual 'ban' from 'b436842423d302bb11cb6f1160d6cb30q9EGL7irEAzdUu1z'",
                    "scope": "Ip",
                    "type": "ban",
                    "value": "18.22.10.20",
                },
                {
                    "duration": "-37m7.335622172s",
                    "id": 97,
                    "origin": "CAPI",
                    "scenario": "crowdsecurity/http-crawl-non_statics",
                    "scope": "Ip",
                    "type": "ban",
                    "value": "185.220.101.204",
                },
            ],
            "new": [
                {
                    "duration": "-37m7.335622172s",
                    "id": 97,
                    "origin": "CAPI",
                    "scenario": "crowdsecurity/http-crawl-non_statics",
                    "scope": "Ip",
                    "type": "ban",
                    "value": "18.22.10.20",
                },
            ],
        }
        self.client.process_response(response)
        assert len(list(self.client.get_new_decision())) == 1
        assert len(list(self.client.get_deleted_decision())) == 2

        assert len(list(self.client.get_new_decision())) == 0
        assert len(list(self.client.get_deleted_decision())) == 0

    def test_empty(self):
        assert self.client.new_decisions.empty() == True
        assert self.client.deleted_decisions.empty() == True

        for _ in self.client.get_deleted_decision():
            pass

        for _ in self.client.get_new_decision():
            pass

    def test_read_write_race(self):
        response = {
            "deleted": [
                {
                    "duration": "-40h37m10.022674981s",
                    "id": 1,
                    "origin": "cscli",
                    "scenario": "manual 'ban' from 'b436842423d302bb11cb6f1160d6cb30q9EGL7irEAzdUu1z'",
                    "scope": "Ip",
                    "type": "ban",
                    "value": str(ipaddress.IPv4Address(v)),
                }
                for v in range(100)
            ],
            "new": [
                {
                    "duration": "-37m7.335622172s",
                    "id": 97,
                    "origin": "CAPI",
                    "scenario": "crowdsecurity/http-crawl-non_statics",
                    "scope": "Ip",
                    "type": "ban",
                    "value": str(ipaddress.IPv4Address(v)),
                }
                for v in range(100)
            ],
        }

        def response_filler():
            for _ in range(100):
                self.client.process_response(response)

        t = threading.Thread(target=response_filler)
        t.start()
        for _ in range(1000):
            list(self.client.get_deleted_decision())
            list(self.client.get_new_decision())



from django.conf import settings
from django.core.exceptions import PermissionDenied
from django.urls import resolve

from pycrowdsec.client import StreamClient


def crowdsec_middleware(get_response):
    def set_settings():
        if not getattr(settings, "PYCROWDSEC_LAPI_KEY"):
            raise Exception("PYCROWDSEC_LAPI_KEY is required")

        settings.pycrowdsec_lapi_url = getattr(
            settings, "PYCROWDSEC_LAPI_URL", "http://localhost:8080/"
        )
        settings.pycrowdsec_user_agent = getattr(
            settings, "PYCROWDSEC_USER_AGENT", "pycrowdsec-django-v1"
        )  # TBD
        settings.pycrowdsec_poll_interval = getattr(settings, "PYCROWDSEC_POLL_INTERVAL", 15)
        settings.pycrowdsec_scopes = getattr(settings, "PYCROWDSEC_SCOPES", ["ip", "range"])
        settings.pycrowdsec_request_transformers = getattr(
            settings,
            "PYCROWDSEC_REQUEST_TRANSFORMERS",
            [lambda request: request.META.get("REMOTE_ADDR")],
        )
        settings.pycrowdsec_actions = getattr(
            settings, "PYCROWDSEC_ACTIONS", {"ban": default_ban_action}
        )

        settings.pycrowdsec_exclude_views = getattr(settings, "PYCROWDSEC_EXCLUDE_VIEWS", set())

    set_settings()
    client = StreamClient(
        api_key=settings.PYCROWDSEC_LAPI_KEY,
        interval=settings.pycrowdsec_poll_interval,
        lapi_url=settings.pycrowdsec_lapi_url,
        scopes=settings.pycrowdsec_scopes,
        user_agent=settings.pycrowdsec_user_agent,
    )

    client.run()

    def middleware(request):
        try:
            if resolve(request.path).view_name in settings.pycrowdsec_exclude_views:
                return get_response(request)
        except:
            return get_response(request)

        for request_transformer in settings.pycrowdsec_request_transformers:
            val = request_transformer(request)
            action = client.get_action_for(val)
            response = get_response(request)
            if action not in settings.pycrowdsec_actions:
                return response
            res = settings.pycrowdsec_actions[action](request)
            if not res:
                return response
            return res

    return middleware


def default_ban_action(request):
    raise PermissionDenied



import logging
import queue
import threading
from abc import ABC, abstractmethod

try:
    from importlib import metadata
except ImportError:  # for Python<3.8
    import importlib_metadata as metadata

__version__ = metadata.version("pycrowdsec")

from time import sleep

import requests

from pycrowdsec.cache import Cache, RedisCache

logger = logging.getLogger(__name__)


def create_session(api_key, insecure_skip_verify, key_path, cert_path, ca_cert_path, user_agent):
    session = requests.Session()
    session.verify = not insecure_skip_verify
    session.headers.update(
        {
            "User-Agent": user_agent,
        },
    )
    if api_key:
        session.headers.update(
            {"X-Api-Key": api_key},
        )
    else:
        if ca_cert_path:
            session.verify = ca_cert_path
        session.cert = (cert_path, key_path)
    return session


class QueryClient:
    def __init__(
        self,
        api_key="",
        lapi_url="http://localhost:8080/",
        user_agent=f"python-bouncer/{metadata.version('pycrowdsec')}",
        insecure_skip_verify=False,
        key_path="",
        cert_path="",
        ca_cert_path="",
    ):
        """
        Initializes a new instance of the CrowdSec API client.

        Args:
            api_key (str): The API key to use for authentication.
            lapi_url (str): The URL of the local API server.
            user_agent (str): The user agent string to use for requests.
            insecure_skip_verify (bool): Whether to skip SSL verification.
            key_path (str): The path to the client's private key file.
            cert_path (str): The path to the client's certificate file.
            ca_cert_path (str): The path to the CA certificate file.
        """

        if api_key == "" and key_path == "" and cert_path == "":
            raise ValueError("You must provide an api_key or a key_path and cert_path")

        self.lapi_url = lapi_url
        self.session = create_session(
            api_key, insecure_skip_verify, key_path, cert_path, ca_cert_path, user_agent
        )

    def get_decisions_for(self, item):
        resp = self.session.get(f"{self.lapi_url}v1/decisions?ip={item}")
        resp.raise_for_status()
        return resp.json()

    def get_action_for(self, item):
        decisions = self.get_decisions_for(item)
        return max(decisions, key=lambda d: d["id"])["type"]


class BaseStreamClient(ABC):
    def __init__(
        self,
        api_key,
        lapi_url="http://localhost:8080/",
        interval=15,
        user_agent=f"python-bouncer/{__version__}",
        scopes=("ip", "range"),
        include_scenarios_containing=(),
        exclude_scenarios_containing=(),
        only_include_decisions_from=(),
        insecure_skip_verify=False,
        key_path="",
        cert_path="",
        ca_cert_path="",
        **kwargs,
    ):
        """
        Initializes a new instance of the CrowdSec API client.

        Args:
            api_key (str): The API key to use for authentication.
            lapi_url (str): The URL of the local API server.
            interval (int): The interval in seconds between each decision request.
            user_agent (str): The user agent string to use for requests.
            scopes (Tuple[str]): The list of scopes to request from the API.
            include_scenarios_containing (Tuple[str]): The list of scenario names to include in decisions.
            exclude_scenarios_containing (Tuple[str]): The list of scenario names to exclude from decisions.
            only_include_decisions_from (Tuple[str]): The list of decision sources to include in decisions.
            insecure_skip_verify (bool): Whether to skip SSL verification.
            key_path (str): The path to the client's private key file.
            cert_path (str): The path to the client's certificate file.
            ca_cert_path (str): The path to the CA certificate file.
            **kwargs: Additional keyword arguments to pass to the requests library.
        """

        if api_key == "" and (key_path == "" and cert_path == ""):
            raise ValueError("You must provide an api_key or a key_path and cert_path")

        self.api_key = api_key
        self.scopes = scopes
        self.interval = int(interval)
        self.lapi_url = lapi_url
        self.user_agent = user_agent
        self.death_reason = None
        self.include_scenarios_containing = include_scenarios_containing
        self.exclude_scenarios_containing = exclude_scenarios_containing
        self.only_include_decisions_from = only_include_decisions_from

        self.session = create_session(
            api_key, insecure_skip_verify, key_path, cert_path, ca_cert_path, user_agent
        )
        self.__post_init__(**kwargs)

    def cycle(self, first_time):
        try:
            url = f"{self.lapi_url}v1/decisions/stream"
            params = {
                "startup": first_time,
                "scopes": ",".join(self.scopes),
                "scenarios_containing": ",".join(self.include_scenarios_containing),
                "scenarios_not_containing": ",".join(self.exclude_scenarios_containing),
                "origins": ",".join(self.only_include_decisions_from),
            }
            for k, v in params.copy().items():
                if not v:
                    del params[k]
            resp = self.session.get(
                url=url,
                params=params,
            )
            resp.raise_for_status()
            self.process_response(resp.json())
        except Exception as e:
            logger.error(f"pycrowdsec got error {e}")
            if first_time == "true":
                self.death_reason = e
                raise e

    def run(self):
        self.cycle("true")  # So we catch errors on startup

        def _thread_cycle():
            while True:
                sleep(self.interval)
                self.cycle("false")

        self.t = threading.Thread(target=_thread_cycle, daemon=True)
        self.t.start()

    def is_running(self):
        return self.t.is_alive()

    @abstractmethod
    def process_response(self, response):
        pass

    @abstractmethod
    def __post_init__(self, **kwargs):
        pass


class StreamClient(BaseStreamClient):
    def __post_init__(self, **kwargs):
        if "redis_connection" in kwargs:
            self.cache = RedisCache(redis_connection=kwargs["redis_connection"])
        else:
            self.cache = Cache()

    def get_action_for(self, item):
        return self.cache.get(item)

    def get_current_decisions(self):
        return self.cache.get_all()

    def process_response(self, response):
        if response["new"] is None:
            response["new"] = []

        if response["deleted"] is None:
            response["deleted"] = []

        for decision in response["deleted"]:
            self.cache.delete(decision["value"])

        for decision in response["new"]:
            self.cache.insert(decision["value"], decision["type"])


class StreamDecisionClient(BaseStreamClient):
    def __post_init__(self, **kwargs):
        self.deleted_decisions = queue.SimpleQueue()
        self.new_decisions = queue.SimpleQueue()

    def get_new_decision(self):
        while not self.new_decisions.empty():
            yield self.new_decisions.get()

    def get_deleted_decision(self):
        while not self.deleted_decisions.empty():
            yield self.deleted_decisions.get()

    def process_response(self, response):
        if response["new"] is None:
            response["new"] = []

        if response["deleted"] is None:
            response["deleted"] = []

        for decision in response["deleted"]:
            self.deleted_decisions.put(decision)

        for decision in response["new"]:
            self.new_decisions.put(decision)



def get_geoip_looker(db_path, scope="city"):
    import geoip2.database
    from geoip2.errors import AddressNotFoundError

    reader = geoip2.database.Reader(db_path)

    def geoip_looker(ip):
        try:
            scope_looker = getattr(reader, scope)
            return scope_looker(ip)
        except AttributeError:
            raise AttributeError(f"The mmdb at {db_path} doesn't support {scope} ")

        except AddressNotFoundError:
            return None

    return geoip_looker



import os

import requests
from expiringdict import ExpiringDict
from flask import Flask, abort, redirect, render_template, request, session, url_for

from pycrowdsec.client import StreamClient
from pycrowdsec.flask import get_crowdsec_middleware

c = StreamClient(
    lapi_url="http://localhost:8080/",
    api_key=os.environ.get("CROWDSEC_LAPI_KEY"),  # your crowdsec LAPI bouncer key goes here
    interval=5,
)
c.run()

app = Flask(__name__)
app.secret_key = "SET_SECRET_KEY"

# In Production env use a "real" cache backend for this. Otherwise
# different app instances would end up having different "valid_captcha_keys".
valid_captcha_keys = ExpiringDict(max_len=1000, max_age_seconds=30)


def validate_captcha_resp(g_recaptcha_response):
    resp = requests.post(
        url="https://www.google.com/recaptcha/api/siteverify",
        data={
            "secret": os.environ.get("GOOGLE_RECAPTCHA_PRIVATE_KEY"),
            "response": g_recaptcha_response,
        },
    ).json()
    return resp["success"]


actions = {
    "ban": lambda: redirect(url_for("ban_page")),
    "captcha": lambda: redirect(url_for("captcha_page"))
    if session.get("captcha_resp") not in valid_captcha_keys
    else None,
}

app.before_request(
    get_crowdsec_middleware(actions, c.cache, exclude_views=("captcha_page", "ban_page"))
)


@app.route("/captcha", methods=["GET", "POST"])
def captcha_page():
    if request.method == "GET":
        return render_template(
            "captcha_page.html", public_key=os.environ.get("GOOGLE_RECAPTCHA_SITE_KEY")
        )
    elif request.method == "POST":
        captcha_resp = request.form.get("g-recaptcha-response")
        if not captcha_resp:
            return redirect(url_for("captcha_page"))

        is_valid = validate_captcha_resp(captcha_resp)
        if not is_valid:
            return redirect(url_for("captcha_page"))

        session["captcha_resp"] = captcha_resp
        valid_captcha_keys[captcha_resp] = None
        return redirect(url_for("index"))


@app.route("/")
def index():
    return "Hello"


@app.route("/ban")
def ban_page():
    return abort(403)


if __name__ == "__main__":
    app.run(host="0.0.0.0")



import random
from unittest import TestCase

from pycrowdsec.cache import Cache


class TestIPv4Cache(TestCase):
    def setUp(self):
        self.cache = Cache()

    def test_insert_delete(self):
        self.cache.insert("1.2.3.4", "ban")
        assert self.cache.get("1.2.3.4") == "ban"
        assert self.cache.get("1.2.3.5") is None

        self.cache.delete("1.2.3.4")
        assert self.cache.get("1.2.3.4") is None
        assert self.cache.get("1.2.3.5") is None

    def test_range_simple_range(self):
        self.cache.insert("0.0.0.0/24", "ban")
        for i in range(256):
            ip = f"0.0.0.{i}"
            with self.subTest(ip=ip):
                assert self.cache.get(ip) == "ban"

        assert self.cache.get(f"0.0.1.0") is None
        assert self.cache.get("::") is None

        self.cache.delete("0.0.0.0/24")
        for i in range(256):
            assert self.cache.get(f"0.0.0.{i}") is None

    def test_range_all_match(self):
        self.cache.insert("0.0.0.0/0", "ban")
        ip_parts = [str(i) for i in range(256)]
        for _ in range(20):
            ip = ".".join(random.choices(ip_parts, k=4))
            with self.subTest(ip=ip):
                assert self.cache.get(ip) == "ban"

        self.cache.delete("0.0.0.0/0")
        for _ in range(20):
            ip = ".".join(random.choices(ip_parts, k=4))
            with self.subTest(ip=ip):
                assert self.cache.get(ip) is None

    def test_range_overlap(self):
        def assert_state(a, b, c):
            assert self.cache.get("0.0.1.2") == a
            assert self.cache.get("0.0.255.255") == b
            self.cache.get("0.1.255.255") == c

        assert_state(None, None, None)
        self.cache.insert("0.0.0.0/16", "ban")
        assert_state("ban", "ban", None)

        self.cache.insert("0.0.0.0/8", "captcha")
        assert_state("ban", "ban", "captcha")

        self.cache.insert("0.0.1.0/24", "throttle")
        assert_state("throttle", "ban", "captcha")

        self.cache.delete("0.0.1.0/24")
        assert_state("ban", "ban", "captcha")

        self.cache.delete("0.0.0.0/8")
        assert_state("ban", "ban", None)

        self.cache.delete("0.0.0.0/16")
        assert_state(None, None, None)


class TestIPv6Cache(TestCase):
    def setUp(self):
        self.cache = Cache()

    def test_insert_delete(self):
        assert self.cache.get("::") is None
        self.cache.insert("::", "ban")
        assert self.cache.get("::") == "ban"
        assert self.cache.get("1::") is None

        self.cache.delete("::")
        assert self.cache.get("1.2.3.4") is None
        assert self.cache.get("1.2.3.5") is None

    def test_range_simple_range(self):
        self.cache.insert("::/112", "ban")
        ips_to_check = random.choices(range(65535), k=100)
        for i in ips_to_check:
            ip = f"::{format(i, 'x')}"
            with self.subTest(ip=ip):
                assert self.cache.get(ip) == "ban"

        assert self.cache.get(f"1::") is None
        assert self.cache.get("0.0.0.0") is None

        self.cache.delete("::/112")
        for i in ips_to_check:
            ip = f"::{format(i, 'x')}"
            with self.subTest(ip=ip):
                assert self.cache.get(ip) is None

    def test_range_all_match(self):
        self.cache.insert("::/0", "ban")
        ip_parts = [format(i, "x") for i in range(65535)]
        for _ in range(20):
            ip = ":".join(random.choices(ip_parts, k=8))
            with self.subTest(ip=ip):
                assert self.cache.get(ip) == "ban"

        self.cache.delete("::/0")
        for _ in range(20):
            ip = ":".join(random.choices(ip_parts, k=8))
            with self.subTest(ip=ip):
                assert self.cache.get(ip) is None

    def test_range_overlap(self):
        def assert_state(a, b, c):
            assert self.cache.get("::ffff") == a
            assert self.cache.get("::1:ffff") == b
            self.cache.get("::1:ffff:ffff") == c

        assert_state(None, None, None)
        self.cache.insert("::/112", "ban")
        assert_state("ban", None, None)

        self.cache.insert("::/98", "captcha")
        assert_state("ban", "captcha", None)

        self.cache.insert("::/82", "throttle")
        assert_state("ban", "captcha", "throttle")

        self.cache.insert("::ffff/128", "block")
        assert_state("block", "captcha", "throttle")

        self.cache.delete("::ffff/128")
        assert_state("ban", "captcha", "throttle")

        self.cache.delete("::/82")
        assert_state("ban", "captcha", None)

        self.cache.delete("::/98")
        assert_state("ban", None, None)

        self.cache.delete("::/112")
        assert_state(None, None, None)


class TestCache(TestCase):
    def setUp(self):
        self.cache = Cache()

    def test_non_ip_items(self):
        assert self.cache.get("TM") is None

        self.cache.insert("TM", "ban")
        assert self.cache.get("TM") == "ban"

        self.cache.insert("TM", "captcha")
        assert self.cache.get("TM") == "captcha"

        self.cache.delete("TM")
        assert self.cache.get("TM") is None

    def test_ip_items(self):
        assert self.cache.get("::ffff") is None

        self.cache.insert("::/112", "ban")
        assert self.cache.get("::ffff") == "ban"
        assert self.cache.get("::abcd") == "ban"
        assert self.cache.get("::abcd:abcd") is None

        assert self.cache.get("1.2.3.4") is None
        self.cache.insert("1.2.0.0/16", "ban")
        assert self.cache.get("1.2.3.4") == "ban"

        self.cache.delete("1.2.0.0/16")
        assert self.cache.get("1.2.3.4") is None

    def test_get_all(self):
        self.cache.insert("CN", "captcha")
        self.cache.insert("1.2.3.4", "ban")
        self.cache.insert("1.2.3.0/24", "ban")
        self.cache.insert("ffff::/24", "ban")

        resp = self.cache.get_all()
        assert resp["ffff::/24"] == "ban"
        assert resp["1.2.3.0/24"] == "ban"
        assert resp["1.2.3.4/32"] == "ban"
        assert resp["CN"] == "captcha"



import pytest
from requests.exceptions import HTTPError
import json

from pycrowdsec.client import StreamDecisionClient


def test_tls_mutual(crowdsec, certs_dir):
    """TLS with two-way bouncer/lapi authentication"""

    lapi_env = {
        "CACERT_FILE": "/etc/ssl/crowdsec/ca.crt",
        "LAPI_CERT_FILE": "/etc/ssl/crowdsec/lapi.crt",
        "LAPI_KEY_FILE": "/etc/ssl/crowdsec/lapi.key",
        "USE_TLS": "true",
        "LOCAL_API_URL": "https://localhost:8080",
    }

    certs = certs_dir(lapi_hostname="lapi")

    volumes = {
        certs: {"bind": "/etc/ssl/crowdsec", "mode": "ro"},
    }

    with crowdsec(environment=lapi_env, volumes=volumes) as cs:
        cs.wait_for_log("*CrowdSec Local API listening*")
        # TODO: wait_for_https
        cs.wait_for_http(8080, "/health", want_status=None)

        port = cs.probe.get_bound_port("8080")
        lapi_url = f"https://localhost:{port}/"

        bouncer = StreamDecisionClient(
            "",
            lapi_url,
            key_path=(certs / "bouncer.key").as_posix(),
            cert_path=(certs / "bouncer.crt").as_posix(),
            ca_cert_path=(certs / "ca.crt").as_posix(),
            user_agent="bouncer_under_test",
        )

        bouncer.cycle("true")
        res = cs.cont.exec_run("cscli bouncers list -o json")
        assert res.exit_code == 0
        bouncers = json.loads(res.output)
        assert len(bouncers) == 1
        assert bouncers[0]["name"].startswith("@")
        assert bouncers[0]["auth_type"] == "tls"
        assert bouncers[0]["type"] == "bouncer_under_test"

        bouncer = StreamDecisionClient(
            "",
            lapi_url,
            key_path=(certs / "agent.key").as_posix(),
            cert_path=(certs / "agent.crt").as_posix(),
            ca_cert_path=(certs / "ca.crt").as_posix(),
        )

        with pytest.raises(HTTPError, match="403"):
            bouncer.cycle("true")

        cs.wait_for_log(
            "*client certificate OU (?agent-ou?) doesn't match expected OU (?bouncer-ou?)*"
        )



import unittest

from redislite import Redis

from pycrowdsec.cache import RedisCache


class TestRedisIntegration(unittest.TestCase):
    def setUp(self):
        self.redis = Redis()
        self.cache = RedisCache(redis_connection=self.redis)

    def test_store_normal_to_redis(self):
        assert not self.redis.exists("pycrowdsec_cache")
        self.cache.insert("124122", "ban")
        assert self.redis.exists("pycrowdsec_cache")
        assert self.redis.hget("pycrowdsec_cache", b"normal_124122") == b"ban"

    def test_ip_to_redis(self):
        self.cache.insert("1.2.3.4", "captcha")
        self.cache.insert("::ffff", "captcha")
        assert self.redis.hgetall("pycrowdsec_cache") == {
            b"ipv4_4294967295_16909060": b"captcha",
            b"ipv6_340282366920938463463374607431768211455_65535": b"captcha",
        }

    def test_delete(self):
        self.cache.insert("1.2.3.4", "captcha")
        self.cache.insert("::ffff", "captcha")
        self.cache.insert("TH", "ban")

        assert self.redis.hlen("pycrowdsec_cache") == 3

        self.cache.delete("TH")
        assert self.redis.hlen("pycrowdsec_cache") == 2
        self.cache.delete("1.2.3.4")

        assert self.redis.hlen("pycrowdsec_cache") == 1
        self.cache.delete("::ffff")

        assert self.redis.hlen("pycrowdsec_cache") == 0

    def test_get_all(self):
        self.cache.insert("1.2.3.4", "captcha")
        self.cache.insert("::ffff", "captcha")
        self.cache.insert("TH", "ban")

        resp = self.cache.get_all()
        assert resp["1.2.3.4/32"] == "captcha"
        assert resp["TH"] == "ban"
        assert resp["::ffff/128"] == "captcha"



import ipaddress
import threading

IPV4_NETMASKS = [int(ipaddress.ip_network(f"0.0.0.0/{i}").netmask) for i in range(32, -1, -1)]

IPV6_NETMASKS = [int(ipaddress.ip_network(f"::/{i}").netmask) for i in range(128, -1, -1)]

NETMASKS_BY_KEY_TYPE = {"ipv4": IPV4_NETMASKS, "ipv6": IPV6_NETMASKS}


def item_to_string(item):
    try:
        ip = ipaddress.ip_network(item)
        return f"ipv{ip.version}_{int(ip.netmask)}_{int(ip.network_address)}"
    except ValueError:
        return f"normal_{item}"


class Cache:
    def __init__(self):
        self.lock = threading.Lock()
        self.cache = {}

    def get(self, item):
        with self.lock:
            key = item_to_string(item)
            key_parts = key.split("_")
            key_type = key_parts[0]
            if key_type == "normal":
                return self.cache.get(key)
            item_network_address = int(key_parts[-1])
            netmasks = NETMASKS_BY_KEY_TYPE[key_type]
            for netmask in netmasks:
                resp = self.cache.get(f"{key_type}_{netmask}_{item_network_address & netmask}")
                if resp:
                    return resp

    def get_all(self):
        with self.lock:
            resp = {}
            for item, action in self.cache.items():
                if item.startswith("normal_"):
                    resp[item.split("_", maxsplit=1)[1]] = action
                elif item.startswith("ipv"):
                    _, netmask, address = item.split("_")
                    ip_network = ipaddress.ip_network((int(address), bin(int(netmask)).count("1")))
                    resp[ip_network.__str__()] = action
            return resp

    def insert(self, item, action):
        key = item_to_string(item)
        with self.lock:
            self.cache[key] = action

    def delete(self, item):
        key = item_to_string(item)
        with self.lock:
            self.cache.pop(key, None)

    def __len__(self):
        with self.lock:
            return len(self.cache)


class RedisCache:
    def __init__(self, redis_connection):
        self.lock = threading.Lock()
        self.redis = redis_connection

    def get(self, item):
        with self.lock:
            key = item_to_string(item)
            key_parts = key.split("_")
            key_type = key_parts[0]
            if key_type == "normal":
                return self.redis.hget("pycrowdsec_cache", key)
            item_network_address = int(key_parts[-1])
            netmasks = NETMASKS_BY_KEY_TYPE[key_type]
            check_for = []
            for netmask in netmasks:
                check_for.append(f"{key_type}_{netmask}_{item_network_address & netmask}")
            responses = self.redis.hmget("pycrowdsec_cache", check_for)
            for response in responses:
                if response:
                    return response.decode()

    def insert(self, item, action):
        with self.lock:
            key = item_to_string(item)
            self.redis.hset("pycrowdsec_cache", key, action)

    def get_all(self):
        with self.lock:
            resp = {}
            for item, action in self.redis.hgetall("pycrowdsec_cache").items():
                item, action = item.decode(), action.decode()
                if item.startswith("normal_"):
                    resp[item.split("_", maxsplit=1)[1]] = action
                elif item.startswith("ipv"):
                    _, netmask, address = item.split("_")
                    ip_network = ipaddress.ip_network((int(address), bin(int(netmask)).count("1")))
                    resp[ip_network.__str__()] = action
            return resp

    def delete(self, item):
        with self.lock:
            key = item_to_string(item)
            self.redis.hdel("pycrowdsec_cache", key)

    def __len__(self):
        with self.lock:
            return self.redis.hlen("pycrowdsec_cache")



from flask import request


def get_crowdsec_middleware(
    actions_by_name,
    crowdsec_cache,
    ip_transformers=[lambda request: request.remote_addr],
    exclude_views=(),
):
    """
    Returns a middleware function for flask, which can be registered by passsing it to app.before_request

        Parameters
        ----------
            actions_by_name(Required):
                Dictionary where key is an action string, eg "ban". The value is a function which carries out the action.
                Eg {"ban": lambda : redirect(url_for("ban_page")) }

            crowdsec_cache(Required):
                An instance of pycrowdsec.cache.Cache.
                It is a KV store, keys being entities like IP, Country strings etc.
                Values are action string for these entities. Eg {"1.2.3.4": "ban"}

            ip_transformers(Optional):
                List of functions which take in the request and produce some other string.
                This produced string is looked up in the cache, if found then then the remediation
                is applied.
                Eg: [lambda ip: ip, lambda ip: get_country_code_for(ip)]

            exclude_views(Optional):
                List of view function names, to exclude crowdsec actions.
                Example: ["ban_view", "captcha_page", "contact_page"]
    """

    def middleware():
        for ip_transformer in ip_transformers:
            action_name = crowdsec_cache.get(ip_transformer(request))
            if not action_name:
                return

            destination_view = request.url_rule.endpoint
            if destination_view in exclude_views:
                return
            if action_name in actions_by_name:
                return actions_by_name[action_name]()

    return middleware



import os

import requests
from django.core.cache import cache
from django.core.exceptions import PermissionDenied
from django.http.response import HttpResponse
from django.shortcuts import redirect, render
from django.urls import reverse

# Create your views here.


def validate_captcha_resp(g_recaptcha_response):
    resp = requests.post(
        url="https://www.google.com/recaptcha/api/siteverify",
        data={
            "secret": os.environ.get("GOOGLE_RECAPTCHA_PRIVATE_KEY"),
            "response": g_recaptcha_response,
        },
    ).json()
    return resp["success"]


def captcha_view(request):
    if not cache.get("valid_captcha_keys"):
        cache.set("valid_captcha_keys", set(), 30)

    if request.method == "GET":
        return render(
            request,
            "captcha_page.html",
            context={"public_key": os.environ.get("GOOGLE_RECAPTCHA_SITE_KEY")},
        )

    elif request.method == "POST":
        captcha_resp = request.POST.get("g-recaptcha-response")
        if not captcha_resp:
            return redirect(reverse("captcha_view"))

        is_valid = validate_captcha_resp(captcha_resp)
        if is_valid:
            valid_keys = cache.get("valid_captcha_keys")
            valid_keys.add(captcha_resp)
            cache.set("valid_captcha_keys", valid_keys, 30)
            request.session["captcha_key"] = captcha_resp
            return redirect(reverse("index"))
        else:
            return redirect(reverse("captcha_view"))


def ban_view(request):
    raise PermissionDenied


def index_view(request):
    return HttpResponse("<h1> Hello  ! </h1>")
